import abc
import typing

from fabricius.types import Data


class Renderer(abc.ABC):
    """
    The Renderer is what translate and generates the output of templates. Core of the work.

    You must subclass this class and override the :py:meth:`render` method, if possible, also add
    a name.
    """

    name: typing.ClassVar[str | None] = None
    """
    The name of the renderer, not necessary, but suggested to add.
    """

    data: Data
    """
    A dictionary that contains data passed by the users to pass inside the template.
    """

    def __init__(self, data: Data) -> None:
        self.data = data

    @abc.abstractmethod
    def render(self, content: str) -> str:
        """
        This method will process a given string, the template input and return the processed
        template as a string too.

        Parameters
        ----------
        content : :py:class:`str`
            The template

        Returns
        -------
        :py:class:`str` :
            The result of the processed template.
        """
        raise NotImplementedError()



import pathlib
import typing

from typing_extensions import Self

from fabricius.app.signals import after_template_commit, before_template_commit
from fabricius.exceptions import (
    AlreadyCommittedError,
    ConflictError,
    MissingRequiredValueError,
)
from fabricius.models.file import File, FileCommitResult
from fabricius.models.renderer import Renderer
from fabricius.types import Data, PathStrOrPath

STATE = typing.Literal["pending", "failed", "persisted"]
RendererType = typing.TypeVar("RendererType", bound=type[Renderer])


class Template(typing.Generic[RendererType]):
    """
    The :py:class:`Template` class represent "a collection of files that, in its whole, represents
    a project"

    The difference between the :py:class:`.Template` class and a collection of :py:class:`.File`
    is that a template assumes all of your files have the same properties. (Requires the same
    renderer, the same data, etc.)

    Typically, a template only use one renderer, and shares the same data across the whole project
    template.

    The :py:class:`.Template` will assist creating a project, while providing a similar interface
    of the :py:class:`.File` model.
    """

    state: STATE
    """
    The state of the template
    """

    base_folder: pathlib.Path
    """
    The folder where the template will be generated.
    """

    files: list[File]
    """
    The list of files that will be rendered when committing.
    """

    data: Data
    """
    The data to pass to the files.
    """

    renderer: RendererType
    """
    The renderer that will be used to generate the files.
    """

    _will_fake: bool

    def __init__(
        self,
        base_folder: PathStrOrPath,
        renderer: RendererType,
    ) -> None:
        """
        Parameters
        ----------
        base_folder : :py:const:`fabricius.types.PathStrOrPath`
            Indication of where the template should be generated.
        renderer : Type of :py:class:`fabricius.models.renderer.Renderer`
            The renderer to use with the template.
        """
        self.base_folder = pathlib.Path(base_folder)
        self.state = "pending"
        self.files = []
        self.data = {}
        self.renderer = renderer
        self._will_fake = False

    @property
    def __files_destinations(self) -> list[pathlib.Path | None]:
        return [file.destination for file in self.files]

    def add_file(self, file: File) -> Self:
        if not file.can_commit:
            reason = file.can_commit
            if reason == "state":
                raise AlreadyCommittedError(file.name)
            raise MissingRequiredValueError(self, reason)

        if file.destination and file.compute_destination() in self.__files_destinations:
            raise ConflictError(
                file,
                f"File {file.name} has a destination that already is present in Template's destinations.",
            )

        self.files.append(file)
        return self

    def add_files(self, files: typing.Iterable[File]) -> Self:
        for file in files:
            self.add_file(file)
        return self

    def push_data(self, data: Data) -> Self:
        self.data = data
        return self

    def fake(self) -> Self:
        self._will_fake = True
        return self

    def restore(self) -> Self:
        self._will_fake = False
        return self

    def commit(self, *, overwrite: bool = False) -> list[FileCommitResult]:
        results: list[FileCommitResult] = []

        before_template_commit.send(self)

        for file in self.files:
            file.with_data(self.data, overwrite=False)
            if self._will_fake:
                file.fake()
            else:
                # Just in case they've been set to fake...
                file.restore()
            result = file.commit(overwrite=overwrite)
            results.append(result)

        after_template_commit.send(self, results)

        return results



import string

from fabricius.models.renderer import Renderer
from fabricius.types import Data

from .utils import DictAllowMiss


class StringTemplateRenderer(Renderer):
    name = "Python string.Template"

    safe: bool
    """
    Indicate if the renderer should use
    :py:meth:`string.Template.safe_substitute` or
    :py:meth:`string.Template.substitute`
    """

    def __init__(self, data: Data, *, safe: bool = True) -> None:
        self.safe = safe
        super().__init__(data)

    def render(self, content: str) -> str:
        if self.safe:
            return string.Template(content).safe_substitute(DictAllowMiss(self.data))
        else:
            return string.Template(content).substitute(self.data)



import typing
from copy import deepcopy
from pathlib import Path

import yaml


class Config(typing.TypedDict):
    default_context: dict[typing.Any, typing.Any]


DEFAULT_CONFIG: Config = {
    "default_context": {},
}


def read_config_file() -> Config:
    try:
        config = Path("~/.cookiecutterrc").expanduser().read_text()
        return yaml.safe_load(config)
    except FileNotFoundError:
        return DEFAULT_CONFIG


def deep_merge(
    base: dict[typing.Any, typing.Any], update_with: dict[typing.Any, typing.Any]
) -> dict[typing.Any, typing.Any]:
    data = deepcopy(base)

    for key, value in update_with.items():
        if isinstance(value, dict):
            data[key] = deep_merge(base.get(key, {}), value)
        else:
            data[key] = value

    return data


def get_config() -> Config:
    conf_data = read_config_file()

    data: Config = deep_merge(DEFAULT_CONFIG, conf_data)  # wtf?

    return Config(default_context=data["default_context"])



from .jinja_renderer import JinjaRenderer as JinjaRenderer
from .mustache import ChevronRenderer as ChevronRenderer
from .python_format import PythonFormatRenderer as PythonFormatRenderer
from .python_string_template import StringTemplateRenderer as StringTemplateRenderer



import errno
import pathlib
import subprocess
import sys
import tempfile
import typing

from fabricius.models.file import FileCommitResult
from fabricius.models.template import Template
from fabricius.readers.cookiecutter.exceptions import FailedHookError
from fabricius.renderers.jinja_renderer import JinjaRenderer
from fabricius.types import Data

HOOKS = ["pre_gen_project", "post_gen_project"]


class AvailableHooks(typing.TypedDict):
    pre_gen_project: typing.Optional[pathlib.Path]
    post_gen_project: typing.Optional[pathlib.Path]


def get_hooks(base_folder: pathlib.Path) -> AvailableHooks | None:
    hooks_folder = base_folder.joinpath("hooks")
    if not hooks_folder.exists():
        return None

    available_hooks: AvailableHooks = {"post_gen_project": None, "pre_gen_project": None}
    for hook_file in hooks_folder.iterdir():
        # annoying mypy moment... requiring literal string...
        if hook_file.stem == "post_gen_project":
            available_hooks["post_gen_project"] = hook_file
        if hook_file.stem == "pre_gen_project":
            available_hooks["pre_gen_project"] = hook_file

    return None if len(available_hooks) == 0 else available_hooks


def run_hook(hook: pathlib.Path, data: Data):
    # Renderer the file
    with tempfile.NamedTemporaryFile(
        delete=False, suffix=hook.suffix, mode="wb"
    ) as temporary_file:
        final_content = JinjaRenderer(data).render(hook.read_text())
        temporary_file.write(final_content.encode("utf-8"))

    path = pathlib.Path(temporary_file.name).resolve()
    cmd = [sys.executable, str(path)] if hook.suffix == ".py" else [str(path)]

    try:
        proc_exit = subprocess.Popen(cmd, shell=sys.platform.startswith("win"), cwd=".").wait(10)
        if proc_exit != 0:
            raise FailedHookError(hook.name, f"Exit status: {proc_exit}")
    except OSError as exception:
        if exception.errno == errno.ENOEXEC:
            raise FailedHookError(
                hook.name, "Might be an empty file or missing a shebang"
            ) from exception
        raise FailedHookError(f"Exception: {exception}") from exception


@typing.overload
def adapt(
    hook: pathlib.Path, type: typing.Literal["pre"]
) -> typing.Callable[[Template[typing.Any]], typing.Any]:
    ...


@typing.overload
def adapt(
    hook: pathlib.Path, type: typing.Literal["post"]
) -> typing.Callable[[Template[typing.Any], list[FileCommitResult]], typing.Any]:
    ...


def adapt(
    hook: pathlib.Path, type: typing.Literal["pre", "post"]
) -> (
    typing.Callable[[Template[typing.Any]], typing.Any]
    | typing.Callable[[Template[typing.Any], list[FileCommitResult]], typing.Any]
):
    if type == "pre":

        def pre_wrapper(template: Template[typing.Any]):
            run_hook(hook, template.data)

        return pre_wrapper

    if type == "post":

        def post_wrapper(template: Template[typing.Any], files_commit: list[FileCommitResult]):
            run_hook(hook, template.data)

        return post_wrapper



import random
import typing

import inflection
from rich.color import Color

FABRICIUS_IS_AWESOME = [
    "Generating your project, hang tight!",
    "Bing, bonk, bang! Scaffolding your next idea...",
    "Helping you on your next idea!",
    "Serious work in progress... Give it a second or two!",
]


def fetch_me_a_beer() -> str:
    return random.choice(FABRICIUS_IS_AWESOME)


def calculate_text_color(
    color: str | Color, *, threshold: int = 150
) -> typing.Literal["black", "bright_white"]:
    """
    Calculate if the text should be in black or white depending on a color.
    It uses the following formula: (r * 0.299 + g * 0.587 + b * 0.114) > threshold
    Where r, g, and b are the RGB (triplet) values of the color.

    Parameters
    ----------
    color : :py:class:`str` or :py:class:`rich.Color`
        The color we're calculating against to.
        This must be a Color class (From Rich)
    threshold : :py:class:`int`
        The threshold to determine if the text is black or white.

    Raises
    ------
    :py:exc:`rich.color.ColorParseError` :
        If ``color`` is of type :py:class:`str`, this will try to parse the color, this error
        will be raised if failed to be parsed.

    Returns
    -------
    str :
        Return either "black" or "bright_white"
    """
    if isinstance(color, str):
        color = Color.parse(color)
    r, g, b = tuple(color.get_truecolor())

    return "black" if (r * 0.299 + g * 0.587 + b * 0.114) > threshold else "bright_white"


def camel_case(text: str) -> str:
    """
    Return the text formatted in camel case

    Parameters
    ----------
    text : :py:class:`str`
        The text you want to format.

    Returns
    -------
    :py:class:`str` :
        The formatted text.

    Example
    -------
    .. code-block:: python

       >>> my_text = "Some text"
       >>> camel_case(my_text)
       'someText'
    """
    return inflection.camelize(text, False)


def snake_case(text: str) -> str:
    """
    Return the text formatted in snake case

    Parameters
    ----------
    text : :py:class:`str`
        The text you want to format.

    Returns
    -------
    :py:class:`str` :
        The formatted text.

    Example
    -------
    .. code-block:: python

       >>> my_text = "Some text"
       >>> snake_case(my_text)
       'some_text'
    """
    return inflection.underscore(text)


def dash_case(text: str) -> str:
    """
    Return the text formatted in dash case

    Parameters
    ----------
    text : :py:class:`str`
        The text you want to format.

    Returns
    -------
    :py:class:`str` :
        The formatted text.

    Example
    -------
    .. code-block:: python

       >>> my_text = "Some text"
       >>> dash_case(my_text)
       'some-text'
    """
    return inflection.dasherize(text)


def pascal_case(text: str) -> str:
    """
    Return the text formatted in pascal case

    Parameters
    ----------
    text : :py:class:`str`
        The text you want to format.

    Returns
    -------
    :py:class:`str` :
        The formatted text.

    Example
    -------
    .. code-block:: python

       >>> my_text = "Some text"
       >>> pascal_case(my_text)
       'SomeText'
    """
    return inflection.camelize(text, True)


def capital_case(text: str) -> str:
    """
    Return the text formatted in capital case

    Parameters
    ----------
    text : :py:class:`str`
        The text you want to format.

    Returns
    -------
    :py:class:`str` :
        The formatted text.

    Example
    -------
    .. code-block:: python

       >>> my_text = "Some text"
       >>> capital_case(my_text)
       'Some Text'
    """
    return inflection.titleize(text)


def sentence_case(text: str) -> str:
    """
    Return the text formatted in sentence case

    Parameters
    ----------
    text : :py:class:`str`
        The text you want to format.

    Returns
    -------
    :py:class:`str` :
        The formatted text.

    Example
    -------
    .. code-block:: python

       >>> my_text = "Some text"
       >>> sentence_case(my_text)
       'Some text'
    """
    has_ending_id = bool(text.endswith("_id"))
    result = inflection.humanize(text)
    if has_ending_id:
        result += " ID"
    return result



# Taken from https://github.com/cookiecutter/cookiecutter/blob/cf81d63bf3d82e1739db73bcbed6f1012890e33e/cookiecutter/extensions.py

import json
import string
import typing
import uuid
from secrets import choice

from jinja2 import Environment
from jinja2.ext import Extension
from slugify import slugify as pyslugify  # type: ignore


class JsonifyExtension(Extension):
    """Jinja2 extension to convert a Python object to JSON."""

    def __init__(self, environment: Environment) -> None:
        """Initialize the extension with the given environment."""
        super().__init__(environment)

        def jsonify(obj: object) -> str:
            return json.dumps(obj, sort_keys=True, indent=4)

        environment.filters["jsonify"] = jsonify  # type: ignore


class RandomStringExtension(Extension):
    """Jinja2 extension to create a random string."""

    def __init__(self, environment: Environment) -> None:
        """Jinja2 Extension Constructor."""
        super().__init__(environment)

        def random_ascii_string(length: int, punctuation: bool = False) -> str:
            if punctuation:
                corpus = "".join((string.ascii_letters, string.punctuation))
            else:
                corpus = string.ascii_letters
            return "".join(choice(corpus) for _ in range(length))

        environment.globals.update(random_ascii_string=random_ascii_string)  # type: ignore


class SlugifyExtension(Extension):
    """Jinja2 Extension to slugify string."""

    def __init__(self, environment: Environment) -> None:
        """Jinja2 Extension constructor."""
        super().__init__(environment)

        def slugify(value: typing.Any, **kwargs: typing.Any) -> str:
            """Slugifies the value."""
            return pyslugify(value, **kwargs)

        environment.filters["slugify"] = slugify  # type: ignore


class UUIDExtension(Extension):
    """Jinja2 Extension to generate uuid4 string."""

    def __init__(self, environment: Environment) -> None:
        """Jinja2 Extension constructor."""
        super().__init__(environment)

        def uuid4() -> str:
            """Generate UUID4."""
            return str(uuid.uuid4())

        environment.globals.update(uuid4=uuid4)  # type: ignore



import contextlib
import typing

_F = typing.ParamSpec("_F")


class Signal(typing.Generic[_F]):
    """
    The Listener is the base class used to create listeners of events.
    """

    listeners: list[typing.Callable[_F, typing.Any]]
    """
    The list of listeners that are subscribed to this signal.
    """

    def __init__(self, *, func_hint: typing.Callable[_F, typing.Any] | None = None) -> None:
        self.listeners = []

    def connect(self, listener: typing.Callable[_F, typing.Any]) -> None:
        """
        Connect a listener to this signal.
        """
        if listener not in self.listeners:
            self.listeners.append(listener)

    def disconnect(self, listener: typing.Callable[_F, typing.Any]) -> None:
        """
        Disconnect a listener to this signal.
        """
        if listener in self.listeners:
            self.listeners.remove(listener)

    def send(self, *args: _F.args, **kwargs: _F.kwargs) -> list[typing.Any]:
        """
        Sends the signal to all subscribed listeners.
        """
        results: list[typing.Any] = []
        for listener in self.listeners:
            with contextlib.suppress(NotImplementedError):
                result = listener(*args, **kwargs)
                results.append(result)
        return results



from fabricius.models.renderer import Renderer

from .utils import DictAllowMiss


class PythonFormatRenderer(Renderer):
    name = "Python str.format"

    def render(self, content: str) -> str:
        return content.format_map(DictAllowMiss(self.data))



import typing
from threading import local

if typing.TYPE_CHECKING:
    from .app.main import App


_local = local()


def get_app() -> "App":
    try:
        return typing.cast("App", _local.app)
    except AttributeError as e:
        raise RuntimeError("App has not been set.") from e


def set_app(app: "App") -> None:
    _local.app = app



import logging
from contextlib import contextmanager
from typing import Any, Generator

from rich.progress import (
    BarColumn,
    Progress,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeRemainingColumn,
)

from fabricius.app.main import logging
from fabricius.app.signals import after_file_commit
from fabricius.models.file import File
from fabricius.types import FileCommitResult

_log = logging.getLogger(__name__)


class TemplateProgressBar:
    total_files: int
    """
    The total files to process.
    Used to be set as the maximum in the progress bar.
    """

    progress: Progress

    task: TaskID | None

    def __init__(self, total_files: int) -> None:
        self.total_files = total_files
        self.progress = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TimeRemainingColumn(),
            transient=True,
        )
        self.task = None

    @contextmanager
    def begin(self, first_message: str) -> Generator[Progress, Any, None]:
        try:
            after_file_commit.connect(self._increase)
            self.task = self.progress.add_task(first_message, total=self.total_files)
            with self.progress as progress:
                yield progress
        finally:
            after_file_commit.disconnect(self._increase)
            self.progress.stop()

    def _increase(self, file: File, result: FileCommitResult) -> None:
        if self.task is None:
            _log.warning("Progress or task not detected. Ignoring.")
            return
        self.progress.update(self.task, advance=1, description=file.name)



class FabriciusError(Exception):
    """
    An error was raised inside Fabricius.

    All exceptions of the Fabricius library **must** subclass this exception.
    """

    def __init__(self, error: object | None = None) -> None:
        super().__init__(error or "Error inside Fabricius. No specific error raised.")


class MissingRequiredValueError(FabriciusError):
    """
    Exception raised when a required value was not set inside an object.
    """

    def __init__(self, instance: object, missing_value: str) -> None:
        """Parameters
        ----------
        instance : object
            The object that holds the missing value.
        missing_value : str
            The value that was not set.
        """
        super().__init__(
            f"{instance.__class__.__name__} is missing a required value to be set: "
            f"'{missing_value}'"
        )


class AlreadyCommittedError(FabriciusError):
    """
    A file has already been committed/persisted.
    """

    def __init__(self, file_name: str) -> None:
        """Parameters
        ----------
        file_name : str
            The file's name that has already been committed.
        """
        super().__init__(f"File '{file_name}' has already been committed.")


class ConflictError(FabriciusError):
    """
    Conflicting parameters between objects.

    This mean that Fabricius cannot continue its work because there would be a conflict between
    one or two objects.

    For example, this error can be raised when two files have the same name in the same
    destination folder.
    """

    def __init__(self, instance: object, reason: str) -> None:
        """
        Parameters
        ----------
        instance : object
            The conflicting object.
        reason : str
            The reason for the conflict.
        """
        super().__init__(f"Conflict error: {instance.__class__.__name__}: {reason}")


class TemplateError(FabriciusError):
    """
    The template has an error that cannot be automatically handled by Fabricius.
    """

    def __init__(self, template_name: str, reason: str) -> None:
        """
        Parameters
        ----------
        template_name : str
            The name of the template that raised the error.
        reason : str
            The reason for the error.
        """
        super().__init__(f"{template_name}: {reason}")



from jinja2 import BaseLoader, Environment

from fabricius.models.renderer import Renderer


class JinjaRenderer(Renderer):
    name = "Jinja Template"

    environment: Environment = Environment(loader=BaseLoader())

    def render(self, content: str) -> str:
        return self.environment.from_string(content).render(**self.data)



import contextlib
import pathlib
import typing

from typing_extensions import Self

from fabricius.app.signals import (
    after_file_commit,
    before_file_commit,
    on_file_commit_fail,
)
from fabricius.exceptions import AlreadyCommittedError, MissingRequiredValueError
from fabricius.models.renderer import Renderer
from fabricius.renderers import (
    ChevronRenderer,
    JinjaRenderer,
    PythonFormatRenderer,
    StringTemplateRenderer,
)
from fabricius.types import FILE_STATE, Data, FileCommitResult, PathStrOrPath


class File:
    """
    The builder class to initialize a file template.
    The result (Through the :py:meth:`.generate` method) is the render of the file's content.
    You can "commit" the file to the disk to persist the file's content.
    """

    name: str
    """
    The name of the file that will be generated.
    """

    state: FILE_STATE
    """
    The state of the file.
    """

    content: str | None
    """
    The template's content.
    """

    template_content: str | None
    """
    The content of the base template, if set.
    """

    destination: pathlib.Path | None
    """
    The destination of the file, if set.
    """

    renderer: type[Renderer]
    """
    The renderer to use to generate the file.
    """

    data: Data
    """
    The data that will be passed to the renderer.
    """

    _will_fake: bool
    """
    If the file should fake its creation upon commit.
    """

    def __init__(self, name: str, extension: typing.Optional[str] = None) -> None:
        """
        Parameters
        ----------
        name : :py:class:`str`
            The name of the file.
        extension : :py:class:`str`
            The extension of the file, without dot, same as ``name="<name>.<extension>"`` (Where
            ``<name>`` and ``<extensions>`` are the arguments given to the class).
        """
        self.name = f"{name}.{extension}" if extension else name
        self.state = "pending"
        self.content = None
        self.destination = None
        self._will_fake = False

        self.renderer = PythonFormatRenderer
        self.data = {}

    def compute_destination(self) -> pathlib.Path:
        """
        Compute the destination of the file.

        Raises
        ------
        :py:exc:`fabricius.exceptions.MissingRequiredValueError` :
            If the object does not have the property "destination" set.
            (Use :py:meth:`.to_directory`)

        Returns
        -------
        pathlib.Path :
            The final path.
        """
        if not self.destination:
            raise MissingRequiredValueError(self, "destination")

        if not self.destination.exists() and (not self._will_fake):
            self.destination.mkdir(parents=True)
        return self.destination.joinpath(self.name)

    @property
    def can_commit(self) -> typing.Literal["destination", "content", "state", True]:
        # sourcery skip: reintroduce-else
        if not self.destination:
            return "destination"
        if not self.content:
            return "content"
        if self.state == "persisted":
            return "state"

        return True

    def from_file(self, path: str | pathlib.Path) -> Self:
        """
        Read the content from a file template.

        Raises
        ------
        :py:exc:`FileNotFoundError` :
            If the file was not found.

        Parameters
        ----------
        path : :py:class:`str` or :py:class:`pathlib.Path`
            The path of the file template.
        """
        path = pathlib.Path(path).resolve()
        self.content = path.read_text()
        return self

    def from_content(self, content: str) -> Self:
        """
        Read the content from a string.

        Parameters
        ----------
        content : :py:class:`str`
            The template you want to format.
        """
        self.content = content
        return self

    def to_directory(self, directory: PathStrOrPath) -> Self:
        """
        Set the directory where the file will be saved.

        Raises
        ------
        :py:exc:`NotADirectory` :
            The given path exists but is not a directory.

        Parameters
        ----------
        directory : :py:class:`str` or :py:class:`pathlib.Path`
            Where the file will be stored. Does not include the file's name.
        """
        path = pathlib.Path(directory).resolve()
        if path.exists() and not path.is_dir():
            raise NotADirectoryError(f"{path} is not a directory.")
        self.destination = path
        return self

    def use_mustache(self) -> Self:
        """
        Use chevron (Mustache) to render the template.
        """
        self.renderer = ChevronRenderer
        return self

    def use_string_template(self) -> Self:
        """
        Use string.Template to render the template.
        """
        self.renderer = StringTemplateRenderer
        return self

    def use_jinja(self) -> Self:
        """
        Use Jinja2 to render the template.
        """
        self.renderer = JinjaRenderer
        return self

    def with_renderer(self, renderer: typing.Type[Renderer]) -> Self:
        """
        Use a custom renderer to render the template.

        Parameters
        ----------
        renderer : Type of :py:class:`fabricius.models.renderer.Renderer`
            The renderer to use to format the file.
            It must be not initialized.
        """
        self.renderer = renderer
        return self

    def with_data(self, data: Data, *, overwrite: bool = True) -> Self:
        """
        Add data to pass to the template.

        Parameters
        ----------
        data : :py:const:`fabricius.types.Data`
            The data you want to pass to the template.
        overwrite : :py:class:`bool`
            If the data that already exists should be deleted. If False, the new data will be
            added on top of the already existing data. Default to ``True``.
        """
        if overwrite:
            self.data = {}
        self.data.update(data)
        return self

    def fake(self) -> Self:
        """
        Set the file to fake the commit.
        This will ensure that the file does not get stored on the machine upon commit.
        """
        self._will_fake = True
        return self

    def restore(self) -> Self:
        """
        Set the file to not fake the commit.
        This will ensure that the file gets stored on the machine upon commit.

        .. hint ::
           This is the default behavior. It's only useful to use this method if you have used :py:meth:`.fake`.
        """
        self._will_fake = False
        return self

    def generate(self) -> str:
        """
        Generate the file's content.

        Raises
        ------
        :py:exc:`fabricius.exceptions.MissingRequiredValue` :
            If no content to the file were added.

        Returns
        -------
        :py:class:`str` :
            The final content of the file.
        """
        if not self.content:
            raise MissingRequiredValueError(self, "content")

        return self.renderer(self.data).render(self.content)

    def commit(self, *, overwrite: bool = False) -> FileCommitResult:
        """
        Save the file to the disk.

        Parameters
        ----------
        overwrite : :py:class:`bool`
            If a file exist at the given path, shall the overwrite parameter say if the file
            should be overwritten or not. Default to ``False``.

        Raises
        ------
        :py:exc:`MissingRequiredValueError <fabricius.exceptions.MissingRequiredValueError>` :
            If a required value was not set. (Content or destination)
        :py:exc:`fabricius.exceptions.AlreadyCommittedError` :
            If the file has already been saved to the disk.
        :py:exc:`FileExistsError` :
            If the file already exists on the disk and ``overwrite`` is set to ``False``.

            This is different from
            :py:exc:`AlreadyCommittedError <fabricius.exceptions.AlreadyCommittedError>`
            because this indicates that the content of the file this generator was never actually
            saved.
        :py:exc:`OSError` :
            The file's name is not valid for the OS.

        Returns
        -------
        :py:class:`fabricius.types.FileCommitResult` :
            A typed dict with information about the created file.
        """
        if not self.destination:
            raise MissingRequiredValueError(self, "destination")
        if not self.content:
            raise MissingRequiredValueError(self, "content")
        if self.state == "persisted":
            raise AlreadyCommittedError(self.name)

        final_content = self.generate()

        destination = self.compute_destination()

        if destination.exists() and not overwrite:
            exception = FileExistsError(f"File '{self.name}' already exists.")
            exception.filename = self.name
            raise exception

        before_file_commit.send(self)

        try:
            if self._will_fake:
                self.state = "persisted"
            else:
                with contextlib.suppress(NotADirectoryError):
                    destination.write_text(final_content)
                    self.state = "persisted"
        except Exception as exception:
            on_file_commit_fail.send(self)

        commit = FileCommitResult(
            name=self.name,
            state=self.state,
            data=self.data,
            template_content=self.content,
            content=final_content,
            destination=self.destination.joinpath(self.name),
            fake=self._will_fake,
        )

        after_file_commit.send(self, commit)
        return commit



from fabricius.exceptions import FabriciusError


class FailedHookError(FabriciusError):
    """
    Raised when a hook run fails.
    """

    exit_code: int | None

    def __init__(
        self, hook: str, reason: str | None = None, *, exit_code: int | None = None
    ) -> None:
        super().__init__(
            f"Hook {hook} failed to run: {reason}"
            if reason
            else f"Hook {hook} has failed to run. No reason provided."
        )
        self.exit_code = exit_code



import typing


class DictAllowMiss(typing.Dict[str, typing.Any]):
    """
    A subclass of Dict to return an empty string in case of missing key, instead of raising an
    error, so that it can play nice with renderer.

    :meta private:
    """

    def __missing__(self, _: str) -> typing.Literal[""]:
        return ""



import json
import pathlib
import sys
import typing
from fnmatch import fnmatch
from functools import partial

from rich import get_console
from rich.prompt import Confirm, Prompt

from fabricius.app.signals import after_template_commit, before_template_commit
from fabricius.app.ui import TemplateProgressBar
from fabricius.exceptions import TemplateError
from fabricius.models.file import File
from fabricius.models.renderer import Renderer
from fabricius.models.template import Template
from fabricius.readers.cookiecutter.config import get_config
from fabricius.readers.cookiecutter.exceptions import FailedHookError
from fabricius.readers.cookiecutter.hooks import adapt, get_hooks
from fabricius.renderers.jinja_renderer import JinjaRenderer
from fabricius.types import FileCommitResult, PathStrOrPath
from fabricius.utils import fetch_me_a_beer, sentence_case

EXTENSIONS = [
    "fabricius.readers.cookiecutter.extensions.JsonifyExtension",
    "fabricius.readers.cookiecutter.extensions.RandomStringExtension",
    "fabricius.readers.cookiecutter.extensions.SlugifyExtension",
    "fabricius.readers.cookiecutter.extensions.UUIDExtension",
    "jinja2_time.TimeExtension",
]


Context = typing.NewType("Context", dict[str, typing.Any])
QuestionContext = typing.NewType("QuestionContext", dict[str, typing.Any])
CookieContext = typing.NewType("CookieContext", dict[str, typing.Any])


class CopyRender(Renderer):
    def render(self, content: str) -> str:
        return content


def obtain_template_path(base_folder: pathlib.Path) -> pathlib.Path | None:
    return next(
        (
            path
            for path in base_folder.iterdir()
            if "cookiecutter" in path.name and "{{" in path.name and "}}" in path.name
        ),
        None,
    )


def wrap_in_cookie(data: Context) -> CookieContext:
    return CookieContext({"cookiecutter": data})


def obtain_files(
    base_folder: pathlib.Path, output_folder: pathlib.Path, data: CookieContext
) -> list[File]:
    files: list[File] = []
    for file_path in base_folder.iterdir():
        if file_path.is_file():
            if "{{" in file_path.name and "}}" in file_path.name:
                file_name = JinjaRenderer(data).render(file_path.name)
            else:
                file_name = file_path.name
            file = File(file_name).from_file(file_path).to_directory(output_folder)
            if should_copy_not_render(file, data):
                file.with_renderer(CopyRender)
            else:
                file.use_jinja()
            files.append(file)
    return files


def should_copy_not_render(file: File, context: CookieContext) -> bool:
    if not context["cookiecutter"].get("_copy_without_render"):
        return False
    to_ignore: list[str] = context["cookiecutter"]["_copy_without_render"]
    for index, value in enumerate(to_ignore):
        # Render the string
        to_ignore[index] = JinjaRenderer(context).render(value)
    return any(fnmatch(str(file.compute_destination()), value) for value in to_ignore)


def read_context_raw(file: pathlib.Path) -> Context:
    if not file.exists():
        raise TemplateError(file.parent.name, f"{file.name} does not exist")
    content = file.read_text()

    try:
        context_data = json.loads(content)
    except json.JSONDecodeError as exception:
        raise TemplateError(
            file.parent.name, f"{file.name} does not appear to be a valid JSON file"
        ) from exception

    return Context(context_data)


def get_questions_only(context: Context) -> QuestionContext:
    question_context = {key: value for key, value in context.items() if not key.startswith("_")}
    return QuestionContext(question_context)


def get_answer(prompts: QuestionContext, *, no_prompt: bool = False) -> dict[str, typing.Any]:
    answers: dict[str, typing.Any] = {}

    for question, default_value in prompts.items():
        if isinstance(default_value, list):
            default_value = typing.cast(list[str], default_value)
            prompt = partial(Prompt(sentence_case(question), choices=default_value))
        else:
            prompt = partial(Prompt(sentence_case(question)), default=default_value)
        answer = "" if no_prompt else prompt()
        answers[question] = answer

    return answers


def setup(
    base_folder: PathStrOrPath,
    output_folder: PathStrOrPath,
    *,
    extra_context: dict[str, typing.Any] | None = None,
    no_prompt: bool = False,
) -> Template[type[JinjaRenderer]]:
    """Setup a template that will be able to be ran once created.

    Parameters
    ----------
    base_folder : :py:const:`PathStrOrPath <fabricius.types.PathStrOrPath>`
        The folder where the template is located. (Choose the folder where the ``cookiecutter.json``
        is located, not the template itself)
    output_folder : :py:const:`PathStrOrPath <fabricius.types.PathStrOrPath>`
        The folder where the template/files will be created once rendered.
    extra_context : :py:const:`Data <fabricius.types.Data>`, optional
        Any extra context to pass to the template.
        It will override the user's prompt.
    no_prompt : bool, optional
        If set to True, no questions will be asked to the user. By default False

    Returns
    -------
    Type of :py:class:`fabricius.models.template.Template`
        The Template that has been generated.
        It is ready to be committed, and everything has been setup.

    Raises
    ------
    :py:exc:`fabricius.exceptions.TemplateError`
        Exception raised when there's an issue with the template that is most probably due to the
        template's misconception.
    """
    if extra_context is None:
        extra_context = {}

    # Obtain the required information first
    base_folder = pathlib.Path(base_folder).resolve()
    output_folder = pathlib.Path(output_folder).resolve()

    # Prepare contexts
    cookiecutter_config_path = base_folder.joinpath("cookiecutter.json")
    user_config = get_config()

    # Ensure a cookiecutter.json file exists.
    # Obtains the context's raw content & the template's hooks.
    if not cookiecutter_config_path.exists():
        raise TemplateError(base_folder.name, "cookiecutter.json does not exist")
    context = read_context_raw(cookiecutter_config_path)
    hooks = get_hooks(base_folder)

    # Obtain the location of the template, if any.
    template_folder = obtain_template_path(base_folder)
    if not template_folder:
        raise TemplateError(base_folder.name, "No template found")

    # Get the template object
    template = Template(template_folder, JinjaRenderer)
    for extension in EXTENSIONS:
        template.renderer.environment.add_extension(extension)
    if context.get("_extensions"):
        for extension in context["_extensions"]:
            template.renderer.environment.add_extension(extension)

    # Add some additional context
    final_context = wrap_in_cookie(context)
    final_context["cookiecutter"] |= {
        "_template": str(template_folder.resolve()),
        "_repo_dir": str(base_folder.resolve()),
        "_output_dir": str(output_folder.resolve()),
    }

    # Begin to get user's prompts.
    questions = get_questions_only(context)
    prompts = get_answer(questions, no_prompt=no_prompt)

    prompts.update(extra_context)

    final_context["cookiecutter"].update(user_config["default_context"])
    final_context["cookiecutter"].update(dict(prompts.items()))

    files = obtain_files(template_folder, output_folder, final_context)
    template.add_files(files)
    template.push_data(final_context)

    if hooks:
        if hook_path := hooks["pre_gen_project"]:
            before_template_commit.connect(adapt(hook_path, "pre"))  # type: ignore
        if hook_path := hooks["post_gen_project"]:
            after_template_commit.connect(adapt(hook_path, "post"))  # type: ignore

    return template


def run(template: Template[type[JinjaRenderer]]) -> list[FileCommitResult]:
    """Run the CookieCutter template generated using :py:func:`.setup`

    Parameters
    ----------
    template : Type of :py:class:`fabricius.models.template.Template`
        The template to render.
    """

    def attempt(force: bool) -> list[FileCommitResult]:
        progress = TemplateProgressBar(len(template.files))
        with progress.begin(fetch_me_a_beer()):
            return template.commit(overwrite=force)

    try:
        return attempt(False)
    except FileExistsError as exception:
        answer = Confirm.ask(
            f"File [cyan]{exception.filename}[/] already exists, this probably means that this template has already been created. Overwrite?"
        )
        if answer:
            return attempt(True)
    except FailedHookError as exception:
        if exception.exit_code:
            sys.exit(exception.exit_code)
        else:
            get_console().print(exception)
            sys.exit(1)
    return []



import pathlib
import typing

if typing.TYPE_CHECKING:
    import os

Data: typing.TypeAlias = "dict[str, typing.Any]"
"""
Data is an alias that represents a dictionary of which every keys are string and their values
of any types.
"""

PathStrOrPath: typing.TypeAlias = "str | os.PathLike[str] | pathlib.Path"
"""
PathStrOrPath represents a path as a :py:class:`str` or a :py:class:`pathlib.Path` object.
Used to help users either have an easy way to give their path.
"""

FILE_STATE = typing.Literal["pending", "persisted"]


class FileCommitResult(typing.TypedDict):
    """
    A FileCommitResult is returned when a file was successfully saved.
    It returns its information after its creation.
    """

    name: str
    """
    The name of the file.
    """

    state: FILE_STATE
    """
    The state of the file. Should always be "persisted".
    """

    destination: pathlib.Path
    """
    Where the file is located/has been saved.
    """

    data: Data
    """
    The data that has been passed during rendering.
    """

    template_content: str
    """
    The original content of the template.
    """

    content: str
    """
    The resulting content of the saved file.
    """

    fake: bool
    """
    If the file was faked.
    If faked, the file has not been saved to the disk.
    """



import chevron

from fabricius.models.renderer import Renderer


class ChevronRenderer(Renderer):
    name = "Chevron (Moustache)"

    def render(self, content: str) -> str:
        return chevron.render(content, self.data)



import typing

from fabricius.models.signal import Signal

if typing.TYPE_CHECKING:
    from fabricius.models.file import File, FileCommitResult
    from fabricius.models.template import Template

    def before_file_commit_hint(file: File):
        ...

    def on_file_commit_fail_hint(file: File):
        ...

    def after_file_commit_hint(file: File, result: FileCommitResult):
        ...

    def before_template_commit_hint(template: Template[typing.Any]):
        ...

    def after_template_commit_hint(
        template: Template[typing.Any], files_commits: list[FileCommitResult]
    ):
        ...

else:
    before_file_commit_hint = None
    on_file_commit_fail_hint = None
    after_file_commit_hint = None
    before_template_commit_hint = None
    after_template_commit_hint = None


before_file_commit = Signal(func_hint=before_file_commit_hint)
"""
A Signal called when a :py:obj:`File <fabricius.models.file.File>` is about to commit a file.
"""

on_file_commit_fail = Signal(func_hint=on_file_commit_fail_hint)
"""
A Signal called when a :py:obj:`File <fabricius.models.file.File>` had an exception occurring
when committing a file.
"""

after_file_commit = Signal(func_hint=after_file_commit_hint)
"""
A Signal called when a :py:obj:`File <fabricius.models.file.File>` has committed a file.
"""

before_template_commit = Signal(func_hint=before_template_commit_hint)
"""
A Signal called when a :py:obj:`Template <fabricius.models.template.Template>` is about to commit
files.
"""

after_template_commit = Signal(func_hint=after_template_commit_hint)
"""
A Signal called when a :py:obj:`Template <fabricius.models.template.Template>` has committed all
the files.
"""
